import click
import inflection
import os


def default(txt):
    return click.style(txt, fg="white", bold=True)


def prompt(txt):
    return click.style(txt, fg="green")


@click.command()
@click.argument("name")
@click.option("--description", prompt=prompt("Description"), default=default("N/A"))
@click.option(
    "--author",
    prompt=prompt("Author name"),
    default=lambda: default(os.environ.get("USER", "")),
)
@click.option(
    "--email",
    prompt=prompt("Author email"),
    default=lambda: default(os.environ.get("USER", "") + "@me.com"),
)
@click.option("--version", prompt=prompt("Version"), default=default("0.0.1"))
@click.option(
    "--django-version", prompt=prompt("Django version"), default=default("1.10")
)
def get_context(name, description, author, email, version, django_version):
    name = click.unstyle(name)
    description = click.unstyle(description)
    email = click.unstyle(email)
    author = click.unstyle(author)
    version = click.unstyle(version)
    django_version = click.unstyle(django_version)
    return {
        "app": inflection.underscore(name),
        "description": description,
        "author": author,
        "email": email,
        "version": version,
        "django_version": django_version,
    }



from __future__ import absolute_import
import click
from dj.application import get_current_application
from dj.utils.system import stdout


@click.command()
def info():
    """Display app info.

    Examples:

    $ dj info
    No application, try running dj init.

    $ dj info
    Application:
      foo @ 2.7.9
    Requirements:
      Django == 1.10

    """
    application = get_current_application()
    info = application.info()
    stdout.write(info)
    return info



from __future__ import absolute_import
import click
from jinja2.exceptions import UndefinedError
from dj.application import get_current_application


class GenerateCommand(click.MultiCommand):
    @property
    def application(self):
        if not hasattr(self, "_application"):
            self._application = get_current_application()
        return self._application

    def list_commands(self, context):
        return self.application.blueprints.keys()

    def get_command(self, context, name):
        return self.application.blueprints[name].load_context()

    def invoke(self, context):
        given_args = context.protected_args + context.args
        interactive = True
        args = []
        for arg in given_args:
            if arg == "--interactive":
                interactive = True
            elif arg == "--not-interactive":
                interactive = False
            else:
                args.append(arg)

        name = args[0]
        application = self.application
        if not application:
            raise click.ClickException("Could not locate application")

        blueprint = application.blueprints.get(name)
        if not blueprint:
            raise click.ClickException("Could not locate blueprint")

        command = blueprint.load_context()
        args = args[1:]
        ctx = command.main(args, standalone_mode=False)

        try:
            return application.generate(blueprint, ctx, interactive=interactive)
        except UndefinedError as e:
            raise click.ClickException(
                "%s.\n"
                "The blueprint's context may be invalid.\n"
                "Blueprint: %s\n"
                "Context: %s" % (str(e), str(blueprint), str(ctx))
            )


@click.command(cls=GenerateCommand)
@click.option("--interactive/--not-interactive", default=True)
def generate(*args, **kwargs):
    """Generate a code stub."""
    pass



import click
import inflection


@click.command()
@click.argument("name")
@click.option("--doc")
def get_context(name, doc):
    """Generate a command with given name.
    The command can be run immediately after generation.

    For example:

        dj generate command bar

        dj run manage.py bar
    """
    name = inflection.underscore(name)
    return {"name": name, "doc": doc or name}



import imp
import os


def load_module(filename):
    """Loads a module from anywhere in the system.

    Does not depend on or modify sys.path.
    """
    path, name = os.path.split(filename)
    name, ext = os.path.splitext(name)
    (file, filename, desc) = imp.find_module(name, [path])

    try:
        return imp.load_module(name, file, filename, desc)
    finally:
        if file:
            file.close()



from __future__ import absolute_import

from dj.utils.system import StyleStdout
import click
from .patch import patch_help_formatting

stdout = StyleStdout()


patch_help_formatting()


class MultiCommand(click.MultiCommand):
    def list_commands(self, ctx):
        return sorted(self.commands.keys())

    def get_command(self, ctx, name):
        # alias-by-prefix support
        all_commands = set(self.list_commands(ctx))
        if name not in all_commands:
            matches = [x for x in all_commands if x.startswith(name)]
            if not matches:
                return None
            elif len(matches) > 1:
                stdout.write("Did you mean one of: %s" % ", ".join(sorted(matches)))
                return None
            else:
                name = matches[0]

        return self.commands.get(name)



from __future__ import absolute_import

import six
import pprint
from redbaron import RedBaron
from redbaron.nodes import (
    AssignmentNode,
    FromImportNode,
    ImportNode,
    DefNode,
    ReturnNode,
    DictNode,
    TupleNode,
    ListNode,
    LiteralyEvaluable
)


def str_format(value):
    return pprint.pformat(value, indent=4)


def merge_lists(source, delta):
    source_set = set(source)
    for d in delta:
        if d not in source_set:
            source.append(d)
    return source


def get_id(node):
    if isinstance(node, AssignmentNode):
        return node.target.dumps()
    if isinstance(node, FromImportNode):
        return node.value.dumps()
    if isinstance(node, ImportNode):
        return node.name.dumps()
    if isinstance(node, DefNode):
        return node.name
    if isinstance(node, ReturnNode):
        return "ReturnNode"
    return None


def merge(source, delta):
    if isinstance(source, six.string_types):
        source = RedBaron(source)
    if isinstance(delta, six.string_types):
        delta = RedBaron(delta)

    if isinstance(source, DictNode) and isinstance(delta, DictNode):
        # two dictionaries
        source_value = source.to_python()
        delta_value = delta.to_python()
        source_value.update(delta_value)
        return RedBaron(str_format(source_value))[0].dumps()

    if isinstance(source, (ListNode, TupleNode)) and isinstance(
        delta, (ListNode, TupleNode)
    ):
        # two lists
        source_value = source.to_python()
        delta_value = delta.to_python()
        value = merge_lists(source_value, delta_value)
        return RedBaron(str_format(value))[0].dumps()

    if isinstance(source, LiteralyEvaluable) and isinstance(delta, LiteralyEvaluable):
        # all other literals -> take new value
        return delta

    elif isinstance(source, FromImportNode) and isinstance(delta, FromImportNode):
        targets = merge_lists(
            [str(x) for x in source.targets], [str(x) for x in delta.targets]
        )
        source.targets = ",".join(targets)
        return source

    if isinstance(source, AssignmentNode) and isinstance(delta, AssignmentNode):
        source.value = merge(source.value, delta.value)
        return source

    if isinstance(source, ImportNode) and isinstance(delta, ImportNode):
        return delta

    if isinstance(source, ReturnNode) and isinstance(delta, ReturnNode):
        return delta

    if hasattr(source, "node_list") and hasattr(delta, "node_list"):
        # two node lists
        for delta_node in delta:
            found = False
            delta_id = get_id(delta_node)
            if delta_id:
                for i, source_node in enumerate(source):
                    source_id = get_id(source_node)
                    if source_node.type == delta_node.type and delta_id == source_id:
                        # found node with matching id
                        # => replace with merged node
                        source[i] = merge(source_node, delta_node)
                        found = True
                        break
            if not found:
                # no matching node, append to list
                source.append(delta_node)
        return source.dumps()

    raise Exception("Can't merge %s into %s" % (delta, source))



from __future__ import absolute_import
import click
from dj.application import get_current_application


@click.argument("addon")
@click.option("--dev", is_flag=True)
@click.command()
def remove(addon, dev):
    """Remove a dependency.

    Examples:

    $ django remove dynamic-rest

    - dynamic-rest == 1.5.0
    """
    application = get_current_application()
    application.remove(addon, dev=dev)



from __future__ import absolute_import
import click
from dj.application import get_current_application
from .run import run


@click.argument("args", nargs=-1, type=click.UNPROCESSED)
@click.command(context_settings={"ignore_unknown_options": True})
def lint(args):
    """Run lint checks using flake8."""
    application = get_current_application()
    if not args:
        args = [application.name, "tests"]
    args = ["flake8"] + list(args)
    run.main(args, standalone_mode=False)



import os
import yaml


class Config(object):

    """A singleton for reading and writing configuration."""

    instances = {}
    defaults = {
        "requirements": "requirements.txt",
        "devRequirements": "requirements.txt.dev",
        "localRequirements": "requirements.txt.local",
        "runtime": "system",
    }
    filename = "dj.yml"
    build_directory = "build"

    def __new__(cls, file):
        if not Config.instances.get(file):
            Config.instances[file] = Config.__Config(file)
        return Config.instances[file]

    def __getattr__(self, name):
        return getattr(self.instance, name)

    def __setattr__(self, name, value):
        return setattr(self.instance, name, value)

    @classmethod
    def get_environment_path(cls, directory):
        return os.path.join(directory, cls.build_directory)

    class __Config(object):
        def __init__(self, directory):
            self.directory = directory
            self.file = os.path.join(directory, Config.filename)
            self._data = {}
            self._load()

        def set(self, key, value):
            self._data[key] = value

        def clear(self, key):
            self._data.pop(key, None)

        def clear_all(self, prefix=""):
            for key in self._data.keys():
                if key.startswith(prefix):
                    self._data.pop(key)

        def get(self, key, default=None):
            return self._data.get(key, default)

        @property
        def environment_path(self):
            return Config.get_environment_path(self.directory)

        def save(self):
            """Save current settings to config file."""
            try:
                self._dump()
            except IOError:
                self._create()
                self._dump()

        def _load(self):
            """Load config file into memory."""
            try:
                with open(self.file, "r") as file:
                    self._data = yaml.load(file)
            except IOError:
                # no config file
                pass
            except yaml.YAMLError as err:
                raise Exception(
                    "Could not parse corrupt config file: %s\n"
                    'Try running "rm %s"' % (str(err), self.file)
                )
            # set defaults
            for key, value in Config.defaults.items():
                if key not in self._data:
                    self._data[key] = value

        def _create(self):
            """Create config file if it does not exist."""
            path = os.path.dirname(self.file)
            if not os.path.exists(path):
                os.makedirs(path)

        def _dump(self):
            with open(self.file, "w") as file:
                yaml.dump(
                    self._data,
                    file,
                    encoding="utf-8",
                    allow_unicode=True,
                    default_flow_style=False,
                )



from collections import OrderedDict
import os
import re
from copy import copy
from dj.utils.system import stdout
from dj.utils.style import white, green, red, yellow, gray


class Dependency(object):
    version_regex = re.compile(r"^([A-Za-z][-A-Za-z0-9_.]+)([=><]+)([0-9.]+)$")
    egg_regex = re.compile(r"^(.*)#egg=(.*)$")

    def __init__(self, value):
        value = value.strip("\n")
        self.value = value
        self.name, self.operator, self.version = Dependency.parse(value)
        self.module_name = self.name.replace("-", "_")

    @classmethod
    def parse(cls, value):
        match = cls.egg_regex.match(value)
        if match:
            # https://some/url@v1#egg=foo
            # -> foo@https://some/url@v1
            name = match.group(2)
            operator = "@"
            version = match.group(1)
        elif os.path.exists(value):
            # /some/local/directory
            name = os.path.basename(value)
            operator = "@"
            version = value
        else:
            # foo>=1.2.0
            match = cls.version_regex.match(value)
            if match:
                name = match.group(1)
                operator = match.group(2)
                version = match.group(3)
            else:
                name = value
                operator = ""
                version = ""
        return name, operator, version

    def __eq__(self, other):
        return str(self) == str(other)

    def __str__(self):
        return self.__unicode__()

    def __unicode__(self):
        return self.value

    def to_stdout(self):
        return (
            "%s %s %s" % (white(self.name), gray(self.operator), green(self.version))
            if self.operator
            else white(self.name)
        )


class DependencyManager(object):
    def __init__(self, source):
        self.source = source

    def add(self, value, warn=True):
        new = Dependency(value)
        new_label = new.to_stdout()
        old = self.dependencies.get(new.name)
        self.dependencies[new.name] = new
        if old is None or str(old) != str(new):
            old_label = old.to_stdout() if old else ""

            self._save()
            if old_label:
                stdout.write(red("- ") + old_label)
            stdout.write(green("+ ") + new_label)
            return True
        else:
            if warn:
                stdout.write("%s %s" % (new_label, yellow("already installed.")))
            return False

    def remove(self, value, warn=True):
        new = Dependency(value)
        old = self.dependencies.get(new.name)
        new_label = new.to_stdout()
        if old is not None:
            old_label = old.to_stdout()
            self.dependencies.pop(new.name)
            self._save()
            stdout.write(red("- ") + old_label)
            return True
        else:
            if warn:
                stdout.write("%s %s" % (new_label, yellow("not installed.")))
            return False

    def get(self, key):
        return self.dependencies.get(key)

    @property
    def dependencies(self):
        if not hasattr(self, "_dependencies"):
            if os.path.exists(self.source):
                self._dependencies = self._load()
            else:
                self._dependencies = {}
        return self._dependencies

    def _load(self):
        dependencies = OrderedDict()
        with open(self.source, "r") as f:
            for line in f.readlines():
                if line:
                    d = Dependency(line)
                    dependencies[d.name] = d
        self._clean = copy(dependencies)
        return dependencies

    def _save(self):
        with open(self.source, "w") as f:
            for dependency in self.dependencies.values():
                dependency = str(dependency)
                if not dependency.endswith("\n"):
                    dependency = dependency + "\n"
                f.write(dependency)
        self._clean = copy(self.dependencies)



from __future__ import absolute_import
import click
from .run import run


@click.argument("args", nargs=-1, type=click.UNPROCESSED)
@click.command(context_settings={"ignore_unknown_options": True})
def test(args):
    """Run tests."""
    arguments = []
    options = []
    for arg in args:
        if arg.startswith("--"):
            options.append(arg)
        else:
            arguments.append(arg)

    if not arguments:
        # by default, run all tests
        arguments = ["tests"]

    args = ["pytest"] + arguments + options
    run.main(args, standalone_mode=False)



import os

from dj.utils.system import execute, exists, stdout, get_platform_os
from dj.utils.style import format_command

PYENV_ROOT = os.path.expanduser("~/.pyenv")


def install_pyenv_dependencies(update=False):
    platform = get_platform_os()
    if platform == "Darwin":
        if not exists("brew"):
            stdout.write(format_command("Installing", "brew"))
            execute(
                'ruby -e "$(curl -fsSL '
                "https://raw.githubusercontent.com"
                "/Homebrew/install/master/install",
                verbose=True,
            )
        elif update:
            stdout.write(format_command("Updating", "brew"))
            execute("brew update", verbose=True)

        if not exists("pyenv"):
            stdout.write(format_command("Installing", "pyenv"))
            execute("brew install pyenv", verbose=True)
        elif update:
            stdout.write(format_command("Updating", "pyenv"))
            execute("brew install --upgrade pyenv", verbose=True)
    elif platform == "Linux":
        if not exists("curl"):
            stdout.write(format_command("Installing", "curl"))
            execute("sudo apt-get install curl")
        if not exists("pyenv"):
            stdout.write(format_command("Installing", "pyenv"))
            execute(
                "curl -L "
                "https://github.com/pyenv/pyenv-installer/"
                "raw/master/bin/pyenv-installer"
                "| bash"
            )
        elif update:
            stdout.write(format_command("Updating", "pyenv"))
            execute("pyenv update", verbose=True)


def get_runtime_root(version):
    if version == "system":
        return None

    return os.path.join(PYENV_ROOT, "versions/%s" % version)


def install_runtime(version):
    if version == "system":
        # use system python version
        if "virtualenv" not in execute("pip list", capture=True):
            execute("pip install -U virtualenv")

    else:
        # use pyenv to manage versions
        install_pyenv_dependencies()
        versions = execute("pyenv versions", capture=True)
        root = get_runtime_root(version)

        if version not in versions:
            stdout.write(format_command("Installing", version))
            execute(
                'PYTHON_CONFIGURE_OPTS="--enable-shared" ' "pyenv install %s" % version,
                verbose=True,
            )

        if "virtualenv" not in execute("%s/bin/pip list" % root, capture=True):
            execute("%s/bin/pip install -U virtualenv" % root)


def get_major_system_version():
    version = execute("python -V", capture=True).split(" ")[-1]
    return get_major_version(version)


def get_major_version(version):
    if version == "system":
        return get_major_system_version()
    return ".".join(version.split(".")[0:2])


def install_virtualenv(virtualenv, directory):
    if not os.path.exists(os.path.join(directory, "bin/activate")):
        # stdout.write(format_command('Creating virtual environment'))
        if not os.path.exists(directory):
            os.makedirs(directory)

        execute("%s %s" % (virtualenv, directory))
        bin_directory = os.path.join(directory, "bin")
        execute("%s/%s" % (bin_directory, "pip install -U pip"))
        execute("%s/%s" % (bin_directory, "pip install -U setuptools"))


class Runtime(object):
    def __init__(self, version, virtual_directory=None, directory=None):
        self.version = version
        self.directory = directory or get_runtime_root(self.version)
        self.virtual_directory = virtual_directory or self.directory

        self._is_virtual = self.virtual_directory != self.directory
        self._is_installed = False

    def _install(self):
        if not self._is_installed:
            if self._is_virtual:
                install_virtualenv(
                    ("%s/bin/virtualenv" % self.directory)
                    if self.directory
                    else "virtualenv",
                    self.virtual_directory,
                )
            else:
                install_runtime(self.version)
            self._is_installed = True

    def execute(self, command, **kwargs):
        self._install()
        env = kwargs.pop("env", {})
        env = " ".join(('%s="%s"' % (k, v) for k, v in env.items()))
        env = "%s " % env if env else ""
        return execute("%s%s/%s" % (env, self.script_directory, command), **kwargs)

    def create_environment(self, directory):
        self._install()
        return Runtime(
            self.version, directory=self.directory, virtual_directory=directory
        )

    @property
    def package_directory(self):
        if not hasattr(self, "_package_directory"):
            self._package_directory = os.path.join(
                self.virtual_directory,
                "lib/python%s/site-packages" % get_major_version(self.version),
            )
        return self._package_directory

    @property
    def script_directory(self):
        if not hasattr(self, "_script_directory"):
            self._script_directory = os.path.join(self.virtual_directory, "bin")
        return self._script_directory

    @property
    def activate_script(self):
        if not hasattr(self, "_activate_script"):
            self._activate_script = os.path.join(self.script_directory, "activate")
        return self._activate_script



from __future__ import absolute_import
import click
from .run import run


@click.command()
@click.argument("port", required=False)
def server(port):
    """Start the Django dev server."""
    args = ["python", "manage.py", "runserver"]
    if port:
        args.append(port)
    run.main(args)



from jinja2 import Environment, StrictUndefined


def strip_extension(string):
    if string.endswith(".j2"):
        string = string[:-3]
    return string


def render_from_string(string, context):
    environment = Environment(undefined=StrictUndefined)
    return environment.from_string(string).render(**context)


def render_from_file(file, context):
    with open(file, "r") as f:
        string = f.read()
    return render_from_string(string, context)



import click


def yellow(message):
    return click.style(message, fg="yellow", bold=True)


def red(message):
    return click.style(message, fg="red", bold=True)


def green(message):
    return click.style(message, fg="green", bold=True)


def gray(message):
    return click.style(message, fg="white", bold=False)


def white(message):
    return click.style(message, fg="white", bold=True)


def prompt(message):
    return click.style(message, fg="green")


def default(message):
    return white(message)


def blue(message):
    return click.style(message, fg="blue", bold=True)


def format_command(a, b="", prefix=""):
    return white(prefix) + blue("%s: " % a) + white(b)



import filecmp
import tempfile
import shutil
import os
import click
from dj.utils.codegen import merge
from dj.utils.jinja import strip_extension, render_from_string, render_from_file
from dj.utils.system import stdout as _stdout
from dj.utils import style


class Generator(object):
    def __init__(self, application, blueprint, context, interactive=True, stdout=None):
        self.stdout = stdout or _stdout
        self.application = application
        self.interactive = interactive
        self.blueprint = blueprint
        self.context = context
        self.temp_dir = tempfile.mkdtemp()

    def __del__(self):
        try:
            shutil.rmtree(self.temp_dir)
        finally:
            self.temp_dir = None

    def generate(self):
        """Generate the blueprint."""
        self.render()
        self.merge()

    def render(self):
        """Render the blueprint into a temp directory using the context."""
        context = self.context
        if "app" not in context:
            context["app"] = self.application.name
        temp_dir = self.temp_dir
        templates_root = self.blueprint.templates_directory
        for root, dirs, files in os.walk(templates_root):
            for directory in dirs:
                directory = os.path.join(root, directory)
                directory = render_from_string(directory, context)
                directory = directory.replace(templates_root, temp_dir, 1)
                os.mkdir(directory)
            for file in files:
                full_file = os.path.join(root, file)
                stat = os.stat(full_file)
                content = render_from_file(full_file, context)
                full_file = strip_extension(render_from_string(full_file, context))
                full_file = full_file.replace(templates_root, temp_dir, 1)
                with open(full_file, "w") as f:
                    f.write(content)
                os.chmod(full_file, stat.st_mode)

    def merge(self):
        """Merges the rendered blueprint into the application."""
        temp_dir = self.temp_dir
        app_dir = self.application.directory
        for root, dirs, files in os.walk(temp_dir):
            for directory in dirs:
                directory = os.path.join(root, directory)
                directory = directory.replace(temp_dir, app_dir, 1)
                try:
                    os.mkdir(directory)
                except OSError:
                    pass
            for file in files:
                source = os.path.join(root, file)
                target = source.replace(temp_dir, app_dir, 1)
                relative_target = target.replace(app_dir, ".")
                action = "r"
                if (
                    os.path.exists(target)
                    and not filecmp.cmp(source, target, shallow=False)
                    and os.stat(target).st_size > 0
                ):
                    # target exists, is not empty, and does not
                    # match source
                    if target.endswith("__init__.py"):
                        # default merge __init__.py files
                        # if non-empty, these should only
                        # contain imports from submoduiles
                        action = "m"
                    elif target.endswith("base.py"):
                        # default skip base.py files
                        # these should be extended by the developer
                        action = "s"
                    else:
                        default = "m"
                        action = (
                            click.prompt(
                                style.prompt(
                                    "%s already exists, "
                                    "[r]eplace, [s]kip, or [m]erge?" % (relative_target)
                                ),
                                default=style.default(default),
                            )
                            if self.interactive
                            else default
                        )
                        action = click.unstyle(action).lower()
                        if action not in {"r", "m", "s"}:
                            action = default

                if action == "s":
                    self.stdout.write(
                        "? %s" % style.white(relative_target), fg="yellow"
                    )
                    continue
                if action == "r":
                    with open(source, "r") as source_file:
                        with open(target, "w") as target_file:
                            target_file.write(source_file.read())
                    self.stdout.write(
                        style.green("+ %s" % style.white(relative_target))
                    )
                if action == "m":
                    with open(target, "r") as target_file:
                        with open(source, "r") as source_file:
                            merged = merge(target_file.read(), source_file.read())
                    with open(target, "w") as target_file:
                        target_file.write(merged)

                    self.stdout.write(
                        style.yellow("> %s" % style.white(relative_target))
                    )



import click
import inflection


@click.command()
@click.argument("name")
def get_context(name):
    """Generate a model with given name.

    You will need to run "dj migrate" to create
    the migration file and apply it to your local
    development database.

    For example:

        dj generate model Foo

        dj migrate
    """
    return {
        "name": inflection.underscore(name),
        "class_name": inflection.camelize(name),
    }



import os
import click
import six
import sys
import subprocess
import signal


def get_platform_os():
    if not exists("uname"):
        return "Windows"

    return execute("uname -s", capture=True)


def true(x):
    return True


def iser(x):
    def inner(y):
        return x == y
    return inner


def get(directory, filter=None, depth=0, include_files=False, include_dirs=False):
    if isinstance(filter, six.string_types):
        flt = iser(filter)
    if not callable(filter):
        # if filter is None/unsupported type, allow all
        flt = true
    else:
        flt = filter

    for root, dirs, files in os.walk(directory):
        search = []
        if include_files:
            search.extend(files)
        if include_dirs:
            search.extend(dirs)

        for file in search:
            file = os.path.join(root, file)
            if flt(file):
                yield file

        depth -= 1
        if depth == 0:
            break


def get_files(directory, filter=None, depth=0):
    return get(directory, filter, depth, include_files=True)


def get_directories(directory, filter=None, depth=0):
    return get(directory, filter, depth, include_dirs=True)


def get_last_touched(file):
    return os.path.getmtime(file) if os.path.exists(file) else None


def touch(file):
    with open(file, "a"):
        os.utime(file, None)


def make_terminate_handler(process, signal=signal.SIGTERM):
    def inner(*args):
        try:
            os.killpg(os.getpgid(process.pid), signal)
        except OSError:
            pass

    return inner


def execute(command, abort=True, capture=False, verbose=False, echo=False, stream=None):
    """Run a command locally.

    Arguments:
        command: a command to execute.
        abort: If True, a non-zero return code will trigger an exception.
        capture: If True, returns the output of the command.
            If False, returns a subprocess result.
        echo: if True, prints the command before executing it.
        verbose: If True, prints the output of the command.
        stream: If set, stdout/stderr will be redirected to the given stream.
            Ignored if `capture` is True.
    """
    stream = stream or sys.stdout

    if echo:
        out = stream
        out.write(u"$ %s" % command)

    # Capture stdout and stderr in the same stream
    command = u"%s 2>&1" % command

    if verbose:
        out = stream
        err = stream
    else:
        out = subprocess.PIPE
        err = subprocess.PIPE

    process = subprocess.Popen(command, shell=True, stdout=out, stderr=err)
    # propagate SIGTERM to all child processes within
    # the process group. this prevents subprocesses from
    # being orphaned when the current process is terminated
    signal.signal(signal.SIGTERM, make_terminate_handler(process))

    # Wait for the process to complete
    stdout, _ = process.communicate()
    stdout = stdout.strip() if stdout else ""
    if hasattr(stdout, "decode"):
        stdout = stdout.decode("utf-8")

    if abort and process.returncode != 0:
        message = u'Error #%d running "%s"%s' % (
            process.returncode,
            command,
            ":\n====================\n" "%s\n" "====================\n" % (stdout)
            if stdout
            else "",
        )
        raise Exception(message)
    if capture:
        return stdout
    else:
        return process


def find_nearest(directory, filename):
    full_path = os.path.join(directory, filename)
    exists = os.path.exists(full_path)
    if exists:
        return full_path
    if directory == "/":
        return None
    return find_nearest(os.path.abspath(os.path.join(directory, os.pardir)), filename)


def exists(program):
    try:
        execute("command -v %s" % program, verbose=False)
        return True
    except Exception:
        return False


class StyleStdout(object):
    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def write(self, message, **kwargs):
        copy = self.kwargs.copy()
        copy.update(kwargs)
        click.echo(click.style(message, **copy))


stdout = StyleStdout(fg="white", bold=True)



from __future__ import absolute_import
import click
from .run import run


@click.command()
def migrate():
    """Run Django migrations.

    This runs "makemigrations" and "migrate".
    """
    run.main(["python", "manage.py", "makemigrations"], standalone_mode=False)
    run.main(["python", "manage.py", "migrate"], standalone_mode=False)



from __future__ import absolute_import
import click
import os
from click.exceptions import ClickException
from dj.application import get_current_application


@click.option("--quiet", is_flag=True, default=False)
@click.argument("args", nargs=-1, type=click.UNPROCESSED)
@click.command(context_settings={"ignore_unknown_options": True})
def run(quiet, args):
    """Run a local command.

    Examples:

    $ django run manage.py runserver

    ...
    """
    if not args:
        raise ClickException("pass a command to run")

    cmd = " ".join(args)
    application = get_current_application()
    name = application.name
    settings = os.environ.get("DJANGO_SETTINGS_MODULE", "%s.settings" % name)
    return application.run(
        cmd,
        verbose=not quiet,
        abort=False,
        capture=True,
        env={"DJANGO_SETTINGS_MODULE": settings},
    )



from __future__ import absolute_import

import os

from .addon import Addon
from .generator import Generator
from .blueprint import Blueprint
from .dependency import DependencyManager, Dependency
from .blueprint import get_core_blueprints
from .utils.system import (
    get_directories,
    get_last_touched,
    find_nearest,
    touch,
    stdout as _stdout,
)
from .utils import style
from .config import Config
from .runtime import Runtime
from redbaron import RedBaron


class Application(object):
    def __init__(self, stdout=None, directory=None):
        self.stdout = stdout or _stdout
        current = os.getcwd()
        nearest_setup_file = find_nearest(current, "setup.py")
        self.directory = directory or (
            os.path.dirname(nearest_setup_file) if nearest_setup_file else current
        )
        self.config = Config(self.directory)
        self.setup_file = os.path.join(self.directory, "setup.py")
        self.requirements_file = os.path.join(
            self.directory, self.config.get("requirements")
        )
        self.dev_requirements_file = os.path.join(
            self.directory, self.config.get("devRequirements")
        )
        self.local_requirements_file = os.path.join(
            self.directory, self.config.get("localRequirements")
        )
        self.runtime = Runtime(self.config.get("runtime"))

    def __str__(self):
        return self.__unicode__()

    def __unicode__(self):
        return "%s (%s)" % (self.name, self.directory)

    @property
    def exists(self):
        if not hasattr(self, "_exists"):
            self._exists = os.path.exists(self.setup_file)
        return self._exists

    @staticmethod
    def parse_application_name(setup_filename):
        """Parse a setup.py file for the name.

        Returns:
            name, or None
        """
        with open(setup_filename, "rt") as setup_file:
            fst = RedBaron(setup_file.read())
            for node in fst:
                if node.type == "atomtrailers" and str(node.name) == "setup":
                    for call in node.call:
                        if str(call.name) == "name":
                            value = call.value
                            if hasattr(value, "to_python"):
                                value = value.to_python()
                            name = str(value)
                            break
                    if name:
                        break
        return name

    def _get_name(self):
        name = self.config.get("name")
        if name:
            return name

        if self.exists:
            try:
                name = Application.parse_application_name(self.setup_file)
            except Exception:
                name = "unknown"
            self.config.set("name", name)
            self.config.save()
        return name

    @property
    def name(self):
        if not hasattr(self, "_name"):
            self._name = self._get_name()
        return self._name

    @property
    def addons(self):
        if not hasattr(self, "_addons"):
            self._addons = {a.name: a for a in self.get_addons()}
        return self._addons

    def get_addons(self):
        self.build()

        addons = []
        for directory in get_directories(
            self.environment.package_directory,
            filter=lambda x: x.endswith("/blueprints"),
        ):
            parent_directory = "/".join(directory.split("/")[0:-1])
            name = os.path.basename(parent_directory)
            addons.append(Addon(name, parent_directory))
        return addons

    def refresh(self):
        if hasattr(self, "_name"):
            del self._name
        if hasattr(self, "_blueprints"):
            del self._blueprints
        if hasattr(self, "_addons"):
            del self._addons
        if hasattr(self, "_exists"):
            del self._exists

    @property
    def blueprints(self):
        if not hasattr(self, "_blueprints"):
            self._blueprints = {}
            for b in self.get_blueprints():
                # add by full name, e.g. dj.model
                self._blueprints[b.full_name] = b
                if not b.addon or b.name not in self._blueprints:
                    # for blueprints other that init or core,
                    # add them to the global namespace
                    self._blueprints[b.name] = b
        return self._blueprints

    def get_blueprints(self):
        addons = self.addons.values()
        blueprints = [a.blueprints.values() for a in addons]
        return get_core_blueprints() + [x for s in blueprints for x in s]

    @property
    def requirements_last_modified(self):
        return get_last_touched(self.requirements_file)

    @property
    def dev_requirements_last_modified(self):
        return get_last_touched(self.dev_requirements_file)

    @property
    def local_requirements_last_modified(self):
        return get_last_touched(self.local_requirements_file)

    @property
    def setup_last_modified(self):
        # timestamp of last setup.py change
        return get_last_touched(self.setup_file)

    @property
    def environment(self):
        if not hasattr(self, "_environment"):
            self._environment = self.runtime.create_environment(
                self.config.environment_path
            )
        return self._environment

    def _get_build_token(self, key):
        return os.path.join(self.environment.virtual_directory, "build.%s" % key)

    def _build(self, key, last_modified, cmd, verbose=True):
        token = self._get_build_token(key)
        last_built = get_last_touched(token)
        if not last_built or last_built < last_modified:
            self.stdout.write(style.format_command("Building", key))
            result = self.execute(cmd, verbose=False, capture=True)
            if "pip" in cmd:
                deps = []
                for line in result.split("\n"):
                    splits = line.split(" ")
                    if line.startswith("Successfully installed"):
                        dep = splits[2]
                        dep = "==".join(dep.rsplit("-", 1))
                        dep = Dependency(dep)
                        deps.append((dep, style.green("+ ")))
                    elif line.startswith("Requirement already satisfied: "):
                        dep = splits[3]
                        dep = Dependency(dep)
                        deps.append((dep, style.yellow(". ")))
                    elif "Uninstalling" in line:
                        index = line.index("Uninstalling")
                        dep = line[index:].split(" ")[1]
                        x = 0
                        dep = dep[x : len(dep) - 1]
                        dep = "".join(dep)
                        dep = "==".join(dep.rsplit("-", 1))
                        dep = Dependency(dep)
                        deps.append((dep, style.red("- ")))

                for dep, prefix in sorted(deps, key=lambda x: str(x[0])):
                    self.stdout.write(prefix + dep.to_stdout())
            touch(token)

    def build(self):
        """Builds the app in the app's environment.

        Only builds if the build is out-of-date and is non-empty.
        Builds in 3 stages: requirements, dev requirements, and app.
        pip is used to install requirements, and setup.py is used to
        install the app itself.

        Raises:
            ValidationError if the app fails to build.
        """

        if self.exists:
            self._build(
                "requirements",
                self.requirements_last_modified,
                "pip install -U -r %s" % self.requirements_file,
            )
            try:
                self._build(
                    "requirements (dev)",
                    self.dev_requirements_last_modified,
                    "pip install -U -r %s" % self.dev_requirements_file,
                )
            except Exception as e:
                if "No such file" not in str(e):
                    raise e
                self.stdout.write(style.yellow("Could not find dev requirements"))
            try:
                self._build(
                    "requirements (local)",
                    self.local_requirements_last_modified,
                    "pip install -U -r %s" % self.local_requirements_file,
                )
            except Exception as e:
                if "No such file" not in str(e):
                    raise e
                self.stdout.write(style.yellow("Could not find local requirements"))
            self._build(
                "application",
                self.setup_last_modified,
                "python %s develop" % self.setup_file,
            )

    def execute(self, command, **kwargs):
        return self.environment.execute(command, **kwargs)

    def run(self, command, **kwargs):
        self.build()
        self.stdout.write(style.format_command("Running", command))
        return self.execute(command, **kwargs)

    def generate(self, blueprint, context, interactive=True):
        """Generate a blueprint within this application."""
        if not isinstance(blueprint, Blueprint):
            bp = self.blueprints.get(blueprint)
            if not bp:
                raise ValueError("%s is not a valid blueprint" % blueprint)
            blueprint = bp

        self.stdout.write(style.format_command("Generating", blueprint.full_name))
        generator = Generator(self, blueprint, context, interactive=interactive)
        result = generator.generate()
        if blueprint.name == "init":
            # try re-setting the name
            self.refresh()
        return result

    def get_dependency_manager(self, dev=False):
        return DependencyManager(
            os.path.join(
                self.directory,
                self.dev_requirements_file if dev else self.requirements_file,
            )
        )

    def add(self, addon, dev=False, interactive=True):
        """Add a new dependency and install it."""
        dependencies = self.get_dependency_manager(dev=dev)
        other_dependencies = self.get_dependency_manager(dev=not dev)
        existing = dependencies.get(addon)
        self.stdout.write(style.format_command("Adding", addon))
        dependencies.add(addon)
        try:
            # try running the build
            self.build()
            self.refresh()

            # remove version of this in other requirements file
            other_dependencies.remove(addon, warn=False)

            # run new addon constructor
            constructor_name = "%s.init" % Dependency(addon).module_name
            constructor = self.blueprints.get(constructor_name)

            if constructor:
                context = constructor.load_context().main([], standalone_mode=False)
                self.generate(constructor, context, interactive=interactive)
        except Exception as e:
            # restore original settings
            self.stdout.write(style.red(str(e)))
            self.stdout.write(style.yellow("Could not find %s" % addon))
            dependencies.remove(addon)
            if existing:
                dependencies.add(existing)
            return

    def remove(self, addon, dev=False):
        """Remove a dependency and uninstall it."""
        dependencies = self.get_dependency_manager(dev=dev)
        other_dependencies = self.get_dependency_manager(dev=not dev)
        self.stdout.write(style.format_command("Removing", addon))
        removed = dependencies.remove(addon, warn=False)
        if not removed:
            removed = other_dependencies.remove(addon, warn=False)

        if removed:
            self.build()
        else:
            exception = "%s is not installed." % Dependency(addon).to_stdout()
            self.stdout.write(style.red(exception))

    def info(self):
        output = []
        dev_requirements = self.get_dependency_manager(dev=True).dependencies
        requirements = self.get_dependency_manager(dev=False).dependencies
        app = self.to_stdout()
        if self.exists:
            output.append(style.blue("Application:\n %s" % app))
            if requirements:
                output.append(style.blue("Requirements:"))
                for _, dep in sorted(requirements.items(), key=lambda x: x[0].lower()):
                    output.append(" " + dep.to_stdout())
            if dev_requirements:
                output.append(style.blue("Requirements (dev):"))
                for _, dep in sorted(
                    dev_requirements.items(), key=lambda x: x[0].lower()
                ):
                    output.append(" " + dep.to_stdout())
        else:
            output.append(
                style.yellow("%s, try running %s." % (app, style.white("dj init")))
            )

        return "\n".join(output)

    def to_stdout(self):
        return (
            "%s %s %s"
            % (
                style.white(self.name),
                style.gray("@"),
                style.green(self.runtime.version),
            )
            if self.name
            else style.yellow("No application")
        )


# singleton application instance
current_application = None


def get_current_application():
    global current_application
    if not current_application:
        current_application = Application()
    return current_application


def set_current_application(application):
    global current_application
    current_application = application



from __future__ import absolute_import
import click

from .base import MultiCommand
from .add import add
from .generate import generate
from .help import help
from .info import info
from .init import init
from .lint import lint
from .migrate import migrate
from .remove import remove
from .run import run
from .server import server
from .shell import shell
from .test import test
from .version import version


class _MultiCommand(MultiCommand):
    commands = {
        "add": add,
        "generate": generate,
        "help": help,
        "info": info,
        "init": init,
        "lint": lint,
        "migrate": migrate,
        "remove": remove,
        "run": run,
        "server": server,
        "shell": shell,
        "test": test,
        "version": version,
    }


@click.command(cls=_MultiCommand)
def dj(*args, **kwargs):
    """DJ, the Django CLI."""
    pass


def execute(command):
    command = command.split(" ")
    return dj.main(command, standalone_mode=False)



from __future__ import absolute_import
import click
import pkg_resources
from dj.utils.system import stdout


@click.command()
def version():
    """Display application version."""
    stdout.write(pkg_resources.get_distribution("djay").version)



import os
from .blueprint import Blueprint


class Addon(object):
    def __init__(self, name, directory):
        self.name = name
        self.module_name = name.replace("-", "_")
        self.directory = directory
        self.blueprints_directory = os.path.join(directory, "blueprints")

    @property
    def blueprints(self):
        if not hasattr(self, "_blueprints"):
            self._blueprints = {b.name: b for b in self.get_blueprints()}
        return self._blueprints

    def get_blueprints(self):
        return Blueprint.get_blueprints(self.blueprints_directory, addon=self)

    def __repr__(self):
        return "%s (%s)" % (self.name, self.directory)



import shutil
import os
from dj.commands.dj import execute
from dj.application import set_current_application, Application
from tempfile import mkdtemp
import multiprocessing


class TemporaryApplication(object):

    DEFAULT_INIT_PARAMS = {
        "app": "dummy",
        "description": "dummy",
        "author": "dummy",
        "email": "dummy@foo.com",
        "version": "0.0.1",
        "django_version": "1.10",
    }

    def __init__(self, params=None):
        self._params = params or self.DEFAULT_INIT_PARAMS
        self._initialized = False
        self._directory = None
        self._application = None

    def _initialize(self):
        if not self._initialized:
            self._initialized = True
            self._directory = mkdtemp()
            self._application = Application(directory=self._directory)
            set_current_application(self._application)
            # generate initial blueprint
            self._application.generate("init", self._params)

    def __del__(self):
        if self._initialized and self._directory:
            shutil.rmtree(self._directory)
            self._initialized = False
            self._directory = None
            self._application = None
            set_current_application(None)

    def execute(self, command, run_async=False):
        def _execute(command):
            cd = os.getcwd()
            try:
                os.chdir(self._directory)
                result = execute(command)
                return result
            finally:
                os.chdir(cd)

        self._initialize()
        if run_async:
            job = multiprocessing.Process(target=_execute, args=(command,))
            job.start()
            return job
        else:
            return _execute(command)



import os
import sys
from dj.utils.system import get_directories, get_files
from dj.utils.imports import load_module


class Blueprint(object):
    @classmethod
    def is_valid(cls, directory):
        # valid if there is a `templates` directory
        # and a `context.py` file
        templates = bool(
            list(
                get_directories(
                    directory, filter=lambda x: x.endswith("/templates"), depth=1
                )
            )
        )
        context = bool(
            list(
                get_files(
                    directory, filter=lambda x: x.endswith("/context.py"), depth=1
                )
            )
        )
        return templates and context

    @classmethod
    def get_blueprints(cls, directory, addon=None):
        return [
            Blueprint(d, addon=addon)
            for d in get_directories(
                directory, filter=lambda x: cls.is_valid(x), depth=1
            )
        ]

    def __init__(self, directory, addon=None):
        self.directory = directory
        self.addon = addon
        self.templates_directory = os.path.join(directory, "templates")
        self.name = os.path.basename(self.directory)
        self.context = os.path.join(directory, "context.py")

    @property
    def full_name(self):
        return "%s.%s" % (self.addon.name, self.name) if self.addon else self.name

    def load_context(self):
        return load_module(self.context).get_context

    def __repr__(self):
        return "%s (%s)" % (self.name, self.directory)


def get_core_blueprints():
    if getattr(sys, "frozen", False):
        # get blueprints relative to sys.executable
        path = os.path.join(os.path.dirname(sys.executable), "dj/blueprints")
    else:
        # get blueprints folder relative to this file
        path = os.path.abspath(os.path.join(os.path.dirname(__file__), "blueprints"))

    blueprints = Blueprint.get_blueprints(path)
    return blueprints



from __future__ import absolute_import
import click
from dj.application import get_current_application


@click.argument("addon")
@click.option("--dev", is_flag=True)
@click.option("--interactive/--not-interactive", is_flag=True, default=True)
@click.command()
def add(addon, dev, interactive):
    """Add a dependency.

    Examples:

    $ django add dynamic-rest==1.5.0

    + dynamic-rest == 1.5.0
    """
    application = get_current_application()
    application.add(addon, dev=dev, interactive=interactive)



from setuptools import (
    find_packages,
    setup,
)


def get_dependencies(file):
    import re

    install_requires = {}
    dependency_links = {}

    url = re.compile(r'[^/]+/.+(?:#egg=(.*))$')
    version = re.compile(r'[0-9][a-zA-Z0-9.]*')

    with open(file) as f:
        for dependency in f.readlines():
            dependency = dependency.strip()
            if dependency.startswith('-e'):
                # ignore interactive mode
                dependency = dependency[3:]

            url_match = url.match(dependency)
            if url_match and url_match.group(1):
                # some/url/dependency#egg=dependency
                egg = url_match.group(1)
                identifier = egg
                separator = egg.rfind('-')
                # check for version
                if separator:
                    egg_name = egg[:separator]
                    egg_version = egg[separator + 1:]
                    if version.match(egg_version):
                        # egg=name-version
                        identifier = '%s==%s' % (egg_name, egg_version)

                install_requires[identifier] = identifier
                dependency_links[identifier] = dependency
            elif url_match:
                # /some/path/to/dependency
                name = dependency.split('/')[-1]
                install_requires[name] = name
                dependency_links[name] = dependency

            else:
                # package==version
                install_requires[dependency] = dependency

    return install_requires, dependency_links


dependencies = get_dependencies('requirements.txt')
install_requires = dependencies[0].values()
dependency_links = dependencies[1].values()

setup(
    author='Anthony Leontiev',
    author_email='alonetiev@gmail.com',
    description='dj: the Django CLI',
    include_package_data=True,
    license=open('LICENSE.md').read().strip(),
    name='djay',
    packages=find_packages(),
    scripts=[],
    install_requires=install_requires,
    dependency_links=dependency_links,
    entry_points={
        'console_scripts': [
            'dj = dj.commands.dj:dj'
        ]
    },
    url='http://github.com/aleontiev/dj',
    version=open('VERSION.txt').read().strip()
)



from __future__ import absolute_import
import os
import click
from dj.config import Config
from dj.utils import style
from .generate import generate
from .base import stdout
from .run import run


@click.command()
@click.argument("name")
@click.option(
    "--runtime",
    prompt=style.prompt("Python version"),
    default=style.default(Config.defaults["runtime"]),
)
def init(name, runtime):
    """Create a new Django app."""
    runtime = click.unstyle(runtime)

    stdout.write(
        style.format_command(
            "Initializing", "%s %s %s" % (name, style.gray("@"), style.green(runtime))
        )
    )

    config = Config(os.getcwd())
    config.set("runtime", runtime)
    config.save()

    generate.main(["init", name], standalone_mode=False)
    run.main(["python", "manage.py", "migrate"])



from __future__ import absolute_import
from click.formatting import HelpFormatter
from dj.utils import style


def patch_help_formatting():
    HelpFormatter._write_usage = HelpFormatter.write_usage
    HelpFormatter.write_usage = _write_usage

    HelpFormatter._write_heading = HelpFormatter.write_heading
    HelpFormatter.write_heading = _write_heading

    HelpFormatter._write_dl = HelpFormatter.write_dl
    HelpFormatter.write_dl = _write_dl


def _write_usage(self, prog, args, prefix="Usage: "):
    prog = style.white(prog)
    args = style.green(args)
    prefix = style.blue(prefix)
    return self._write_usage(prog, args, prefix=prefix)


def _write_heading(self, heading):
    return self._write_heading(style.blue(heading))


def _write_dl(self, rows, **kwargs):
    return self._write_dl([(style.green(r[0]), r[1]) for r in rows], **kwargs)



from __future__ import print_function
from unittest import TestCase
import click
from dj.utils.style import yellow, white
from dj.test import TemporaryApplication
import time
import requests


class CLITestCase(TestCase):

    def log(self, msg):
        print(white('\n*** ') + yellow(msg))

    def test_create_new_app(self):
        application = TemporaryApplication()

        self.log('Testing build')
        result = application.execute('run python manage.py help --quiet')
        self.assertTrue('for help on a specific subcommand' in result, result)

        self.log('Testing lint')
        result = application.execute('run flake8 dummy --quiet')
        self.assertFalse(result)

        self.log('Testing migrate')
        result = application.execute('run python manage.py migrate --quiet')
        self.assertTrue('Applying auth.0001_initial' in result, result)

        result = application.execute('run manage.py migrate --quiet')
        self.assertFalse('Applying auth.0001_initial' in result, result)

        self.log('Testing generate model')
        application.execute('generate model foo --not-interactive')

        # testing fails
        result = application.execute('run pytest --quiet')
        self.assertTrue('no such table: dummy_foo' in result, result)

        application.execute('run manage.py makemigrations --quiet')

        result = application.execute('run manage.py migrate --quiet')
        self.assertTrue('Applying dummy.0001_initial' in result, result)

        result = application.execute('run manage.py migrate --quiet')
        self.assertFalse('Applying dummy.0001_initial' in result, result)

        self.log('Testing test')
        result = application.execute('run pytest --quiet')
        self.assertTrue('1 passed' in result)

        self.log('Testing generate command')
        application.execute('generate command foobar')

        result = application.execute('run manage.py help --quiet')
        self.assertTrue('foobar' in result, result)

        result = application.execute('run manage.py foobar --quiet')
        self.assertTrue('foobar called' in result, result)

        self.log('Testing info')
        result = click.unstyle(application.execute('info'))
        self.assertTrue('Django == 1.10' in result, result)

        self.log('Testing serve')
        server = application.execute(
            'run manage.py runserver 9123 --quiet', run_async=True
        )
        time.sleep(5)

        response = requests.get('http://localhost:9123')
        content = response.content
        self.assertTrue('It worked' in content.decode('utf-8'), content)

        server.terminate()

        self.log('Testing version')
        application.execute('version')
